#device PIC16F877 
#include "..\16f877.h"
#fuses HS, NOWDT, NOPROTECT, PUT, BROWNOUT
#use delay(clock=20000000)

#define DELAY							45							// 5 sec @ 20MHz, reduce for lower frequencies
#define START_HIGH				0x0C						// Make this 0x0C for 4K memory, 0x1C for 8K
#define START_LOW					0xE0						// This one stays the same

#define TO_MAIN						0x50						// Address of main() above start_h, start_l

#define FLASH_LOADED			0x80
#define ERROR 						0x40
#define USER_CODE_LOADED 	0x20
#define QUICK_START				0x10

#define BLINKER						pin_C0

#use fast_io(C)
#use rs232(baud = 19200, xmit=pin_c6, rcv=pin_c7)

																					// Amazing that this works
#org (START_HIGH << 8) + START_LOW, (START_HIGH << 8) + START_LOW + 0x0F
void start(void)
{	
	#asm	
		MOVLW (START_HIGH & 0x18)			// Jump to the loader's Main routine. This
		MOVWF 0x0A			// gets placed here by the loader, but we expect it to be
		GOTO ((START_HIGH << 8) + START_LOW + TO_MAIN) & 0x7FF
		NOP							// replaced by a jump to the user's code.

		MOVLW 0x0				// These 3 words deal with the situation where the user's
		MOVWF 0x0A			// first 4 words of code aren't a jump, which means 4 is
		GOTO 0x04				// the next address to be executed. So we just run through
	#endasm						// the code meant to be at 0-3 and then do a jump to 4.

										// The compiler puts some stuff after the #endasm but it'll
}										// never execute, so who cares.


#org (START_HIGH << 8) + START_LOW + 0x10, (START_HIGH << 8) + START_LOW + 0x4F
void flash_load( int select )
{
	int data_l, data_h, i;

	data_l = eedata;
	data_h = eedath;
	i = 2;

	do
	{	if (select)														// Nonzero says load to flash
			bit_set(eepgnd);										// Else to EEPROM
		else
			bit_clear(eepgnd);

		bit_set(wren);
		eecon2 = 0x55;
		eecon2 = 0xAA;
		bit_set(wr);													// Do the deed
		#asm
			nop
			nop																	// Skip 2 instructions
		#endasm
		bit_clear(wren);					 

		bit_set(rd);													// Read data back
		#asm																
			nop																	// Skip 2 instructions
			nop
		#endasm

		if ( (data_l == eedata) && (data_h == eedath) ) // Matched?
			break;															// Yes, all done

		eedata = data_l;											// Ready to load same again
		eedath = data_h;

	}	while (--i > 0 );					// Failed, repeat a max of once more
}				

int loads[8] = {START_HIGH & 0x18, 0x30, 0x8A, 0x00, 
	START_LOW + TO_MAIN, 0x28 + ((((START_LOW + TO_MAIN) > 0xFF? 1: 0) + START_HIGH) & 7), 0, 0};		

#org (START_HIGH << 8) + START_LOW + TO_MAIN, (START_HIGH << 8) | 0x3FF				// Give it all remaining ROM

void main()
{	int nums[40];
	int startup[8];
	int z, nb, addr_l, addr_h, cs, su_count;
	int clox, t1, t2, t3, phase;

#ifdef BLINKER
	set_tris_c(0xBE);										// All inputs except bit 6 (TX) and 0 (BLINKER)
	output_bit(BLINKER, 1 );						// Light ON (loader has control)
#else
	set_tris_c(0xBF);										// All inputs except bit 6 (TX)
#endif

	su_count = 0;
SA:
	disable_interrupts(GLOBAL);

	eeadrh = START_HIGH;
	eeadr = START_LOW;

	for (clox = 0; clox < 8; clox += 2)
	{
		bit_set(eepgnd);								// Grab 2-byte words from flash
		bit_set(rd);										// Has user startup code been loaded?
		#asm
			nop
			nop
		#endasm
		if (eedata != loads[clox] || eedath != loads[clox+1])
			su_count |= USER_CODE_LOADED;								// Yes if any data in 0-3 changed
		eeadr++;												// Assume that if any code's there, it's valid
	}

	while (1)
	{
		phase = 0;
		cs = 0;
		clox = 0;																// Start of a line
		
		while(1)
		{
			t1 = 0;																// Loop within a line
			t2 = 0;
			t3 = (su_count & QUICK_START? 5: DELAY);

			do
			{
				restart_wdt();
				if ((su_count & USER_CODE_LOADED) && (--t1 == 0))
				{
					
					if (--t2 == 0)									// Delay about 5 seconds
					{	
						if (--t3 == 0)
						{
							#ifdef BLINKER
								output_bit(BLINKER, 0 );						// Light OFF (user has control)
							#endif
							start();													// Jump to user code
						}
						#ifdef BLINKER
							output_bit(BLINKER, (t3 & 1) );			// Flash light (loader has control)
						#endif
					}																			
				}	
			}	while(!bit_test(rcif));									// Wait for character

			z = rcreg;																// Read it

			if (z == ':')															// Colon at start, line
			{																					// has begun
				clox = 0;
				phase = 1;
				cs = 0;
				continue;
			}

			else if (clox > 35)												// Line too long
			{	phase = 0x81;														// Highest index is 35
				continue;																// but it should break before
			}																					// reaching here
			
			else if (!( (z >= 'a' && z <= 'f') || (z >= 'A' && z <= 'F') ||
			 (z >= '0' && z <= '9')))
			{	if (phase)															// Set flag for
					phase = 0x82;													// bad character in a line
				continue;																// Ignore if not in a line
			}

			if (phase == 1)														// First character of pair
			{	if (z >= '0' && z <= '9')								// High nibble
					nums[clox] = (z - '0') << 4;
				else
					nums[clox] = (toupper(z) - 'A' + 10) << 4;
				phase = 2;
			}
			
			else if (phase == 2)											// Second character of pair
			{	if (z >= '0' && z <= '9')								// Low nibble
					nums[clox] += (z - '0');
				else
					nums[clox] += (toupper(z) - 'A' + 10);

				cs -= nums[clox];												// Checksum should end up 0
				phase = 1;

				if (clox == nums[0] + 4)								// Seen a whole line
					break;

				clox++;
			}			
		}

		if (cs != 0 && phase < 0x80)								// Bad checksum
			phase = 0x83;

		if (phase >= 0x80)													// Error, don't allow startup
		{	su_count |= ERROR;
			continue;
		}
		
		if (nums[3] == 0)														// Load data to flash
		{
			cs = nums[1]/2;														// Cs now holds upper byte of addr to load
			eeadrh = cs;
			eeadr = nums[2]/2 + (nums[1] & 1? 0x80: 0);			

			for (clox = 0; clox < nums[0]; clox += 2)	// Load the data to flash
			{																					// Note that each word is 2 bytes
				if (cs == 0 && eeadr < 4)								// First 4 words go to RAM
				{
					startup[nums[2] + clox] = nums[nums[2]+ clox + 4];
					startup[nums[2] + clox + 1] = nums[nums[2] + clox + 5];
					su_count |= (1 << eeadr);							// Should finish up with 00001111
				}

				else																		// If not RAM, then flash or EEPROM
				{
					eedata = nums[clox + 4];
					eedath = nums[clox + 5];
					
					if (cs == 0x21)												// Load to EEPROM?
					{
						while (bit_test(wr)) ;							// Wait for previous cycle to finish
						flash_load(0);											// EEPROM data goes to 0x21xx
						continue;
					}

					else if (cs > START_HIGH)							// Address is too high
						continue;

					else if ( (cs == START_HIGH) && (eeadr >= START_LOW) )
						continue;														// Don't load anything
					
					flash_load(1);
					su_count |= FLASH_LOADED;							// Flag says new data loaded
					restart_wdt();
				}

				if (++eeadr == 0)											// Increment address
					eeadrh++;
				
			}	// End FOR loop loading one line's data to RAM, flash or EEPROM

			if (su_count & (USER_CODE_LOADED + FLASH_LOADED) == USER_CODE_LOADED + FLASH_LOADED)
			{	su_count &= ~USER_CODE_LOADED;				// Can't run user code now
				eeadrh = START_HIGH;									// No need to load this if it was 
				eeadr = START_LOW;										// found in ROM at the start 
				for (clox = 0; clox < 8; clox += 2)		// as determined by state of U_C_L
				{
					eedata = loads[clox];
					eedath = loads[clox+1];
					flash_load(1);
					eeadr++;
					restart_wdt();
				}
			}

			cs = nums[1]/2;														// Starting address of line in flash
			eeadrh = cs;
			eeadr = nums[2]/2 + (nums[1] & 1? 0x80: 0);			
			cs = 0;

			for (clox = 0; clox < nums[0]; clox += 2)	// Read the data from flash
			{
				if (eeadrh == 0 && eeadr < 4)						// First 4 words come from RAM
				{
					cs += startup[nums[2] + clox];
					cs += startup[nums[2] + clox + 1];
				}

				else																		// If not RAM, then flash or EEPROM
				{
					if (eeadrh == 0x21)
					{
						bit_clear(eepgnd);									// Addresses 0x21xx call for data from EEPROM
						eedath = 0;													// Set upper byte to 0 for 8-bit read
					}
					else
						bit_set(eepgnd);										// Else from flash

					bit_set(rd);													// Read data back
					#asm
						nop
						nop
					#endasm

					cs += eedata;
					cs += eedath;
				}
				if (++eeadr == 0)												// Inc low address
					eeadrh++;															// Maybe high address too
			}

			txreg = cs;																// Transmit checksum

			#ifdef BLINKER
				if (input(BLINKER))
					output_bit( BLINKER, 0 );
				else
					output_bit( BLINKER, 1 );								// Alternate LED states each line
			#endif
		
		}	// End load to flash for a single line

		else if (nums[3] == 1)											// Start execution
		{											// Was startup address loaded, and was data error-free?
			if ((su_count & (ERROR + 7)) == 7)
			{																		// Yes, load user's startup code
				eeadrh = START_HIGH;
				eeadr = START_LOW;
				
				for (clox = 0; clox < 8; clox += 2)
				{
					eedata = startup[clox];
					eedath = startup[clox+1];

					flash_load(1);
					restart_wdt();
					eeadr++;
				}
				while(!bit_test(trmt)) ;
					txreg = 0x55;									// Send code to user to indicate startup
				while(!bit_test(trmt)) ;

				su_count = QUICK_START;
				goto SA;
			}

			while(!bit_test(trmt)) ;
				txreg = su_count;									// Send code to user to indicate failed startup
			su_count = 0;

		}							
	}
}


